MODULE GraphLib;

IMPORT SDL := GraphSdl, (*Mix := SDL2mixer, Net := SDL2net,*)
       SYSTEM, S := Strings, StdLog := Out;

CONST
  (* Flip Flags *)
  flipNone* = {};
  flipH*    = 0;
  flipV*    = 1;
  flipHV*   = {flipH, flipV};

  (* Draw Mode Flags *)
  drawSpriteNormal* = 0;
  drawSpriteLit*    = 1;
  drawSpriteTrans*  = 2;

  (* Settings *)
  fullscreen* = 0;
  spread*     = 2;
  initMouse*  = 8;
  initSound*  = 9;
  initNet*    = 10;

  (* Mouse Buttons *)
  btnLeft*  = 0;
  btnRight* = 1;
  btnMid*   = 2;

  (* Key Codes *)
  keyA*           = 4;
  keyB*           = 5;
  keyC*           = 6;
  keyD*           = 7;
  keyE*           = 8;
  keyF*           = 9;
  keyG*           = 10;
  keyH*           = 11;
  keyI*           = 12;
  keyJ*           = 13;
  keyK*           = 14;
  keyL*           = 15;
  keyM*           = 16;
  keyN*           = 17;
  keyO*           = 18;
  keyP*           = 19;
  keyQ*           = 20;
  keyR*           = 21;
  keyS*           = 22;
  keyT*           = 23;
  keyU*           = 24;
  keyV*           = 25;
  keyW*           = 26;
  keyX*           = 27;
  keyY*           = 28;
  keyZ*           = 29;
  key1*           = 30;
  key2*           = 31;
  key3*           = 32;
  key4*           = 33;
  key5*           = 34;
  key6*           = 35;
  key7*           = 36;
  key8*           = 37;
  key9*           = 38;
  key0*           = 39;
  key1Pad*        = 89;
  key2Pad*        = 90;
  key3Pad*        = 91;
  key4Pad*        = 92;
  key5Pad*        = 93;
  key6Pad*        = 94;
  key7Pad*        = 95;
  key8Pad*        = 96;
  key9Pad*        = 97;
  key0Pad*        = 98;
  keyF1*          = 58;
  keyF2*          = 59;
  keyF3*          = 60;
  keyF4*          = 61;
  keyF5*          = 62;
  keyF6*          = 63;
  keyF7*          = 64;
  keyF8*          = 65;
  keyF9*          = 66;
  keyF10*         = 67;
  keyF11*         = 68;
  keyF12*         = 69;
  keyEsc*         = 41;
  keyTilde*       = 53;
  keyMinus*       = 45;
  keyEquals*      = 46;
  keyBackspace*   = 42;
  keyTab*         = 43;
  keyOpenBrace*   = 47;
  keyCloseBrace*  = 48;
  keyEnter*       = 40;
  keyColon*       = 51;
  keyQuote*       = 52;
  keyBackslash*   = 49;
  keyBackslash2*  = 100;
  keyComma*       = 54;
  keyStop*        = 55;
  keySlash*       = 56;
  keySpace*       = 44;
  keyInsert*      = 73;
  keyDel*         = 76;
  keyHome*        = 74;
  keyEnd*         = 77;
  keyPgUp*        = 75;
  keyPgDn*        = 78;
  keyLeft*        = 80;
  keyRight*       = 79;
  keyUp*          = 82;
  keyDown*        = 81;
  keySlashPad*    = 84;
  keyAsterisk*    = 85;
  keyMinusPad*    = 86;
  keyPlusPad*     = 87;
  keyDelPad*      = 99;
  keyEnterPad*    = 88;
  keyPrtScr*      = 70;
  keyPause*       = 72;

  keyMODIFIERS*   = 115;

  keyLShift*      = 115;
  keyRShift*      = 116;
  keyLControl*    = 117;
  keyRControl*    = 118;
  keyAlt*         = 119;
  keyAltGr*       = 120;
  keyLWin*        = 121;
  keyRWin*        = 122;
  keyMenu*        = 123;
  keyScrLock*     = 124;
  keyNumLock*     = 125;
  keyCapsLock*    = 126;

  keyMax*         = 127;

TYPE
  Pixels* = SDL.Pixels;

  Bitmap* = POINTER TO RECORD
    surface*: SDL.Surface;
    w*, h*: INTEGER
  END;

  Font* = POINTER TO RECORD
    bmp*: Bitmap;
    charW*, charH*: INTEGER;
    charRows*, charsInRow*: INTEGER
  END;

  KeyArray* = SDL.KeyArray;

(*
  Sample* = POINTER TO RECORD
    chunk*: Mix.Chunk
  END;

  Music* = POINTER TO RECORD
    music*: Mix.Music
  END;
*)
  Region* = RECORD
    x*, y*, w*, h*: INTEGER
  END;

  (* SDL Net *)
(*  
  IPAddress* = Net.IPAddress;
  Socket* = Net.TCPSocket;

  SocketSet* = Net.SocketSet;
*)
  PStr* = POINTER TO ARRAY OF SHORTCHAR;

(*
  NetBuf* = POINTER TO RECORD
    s: PStr;
    len: INTEGER; (* Actual used length of the buffer *)
    lastLen: INTEGER
  END;
*)

VAR
  window: SDL.Window;
  renderer: SDL.Renderer;
  screen-: Bitmap;

  CloseButtonCallback: PROCEDURE;

  keyBuf: ARRAY 128 OF INTEGER;
  keyLen: INTEGER; (* Amount of keys in keyBuf *)

  wantW, wantH: INTEGER;
  wantFullscreen, wantSpread: BOOLEAN;
  wantMouse, wantSound, wantNet: BOOLEAN;
  closing-: BOOLEAN; (* True when window is being closed by user *)

  (* Flip Region *)
  flipRegion: Region;

  (* Mouse *)
  mouseX, mouseY: INTEGER;
  mouseFocusX, mouseFocusY: INTEGER;
  lastBlitMouseOutside: BOOLEAN;
  lastBlitMouseX, lastBlitMouseY: INTEGER;
  needRedrawMouse: BOOLEAN; (* True if mouse has moved since last redraw *)
  showMouse: BOOLEAN; (* Whether to show mouse pointer on screen *)
  mousePointer: Bitmap;
  underMouse: Bitmap; (* Buffer to copy part of image under the mouse *)

(* C Macros *)

PROCEDURE LoadBmp(fname: ARRAY OF SHORTCHAR): SDL.Surface;
BEGIN
  RETURN SDL.LoadBmpRW(SDL.RWFromFile(fname, 'rb'), 1)
END LoadBmp;

(* Misc *)

(*
PROCEDURE Str*(n: INTEGER; VAR s: ARRAY OF SHORTCHAR);
BEGIN
  IntStr.IntToStr(n, s)
END Str;

PROCEDURE Val*(s: ARRAY OF SHORTCHAR): INTEGER;
VAR n: INTEGER; res: IntStr.ConvResults;
BEGIN
  IntStr.StrToInt(s, n, res);
  IF res # IntStr.strAllRight THEN n := 0 END;
  RETURN n
END Val;

PROCEDURE AppendN*(n: INTEGER; VAR s: ARRAY OF SHORTCHAR);
VAR sn: ARRAY 30 OF CHAR;
BEGIN
  Str(n, sn); S.Append(sn, s)
END AppendN;
*)

(* General *)

PROCEDURE GetError*(VAR s: ARRAY OF SHORTCHAR);
VAR p: SDL.PStr; i: INTEGER;
BEGIN
  p := SDL.GetError();
  s := p$
  (*
  i := 0;
  WHILE (i < LEN(s)) & (p[i] # 0X) DO s[i] := p[i]; INC(i) END*)
END GetError;

PROCEDURE StdCloseButtonCallback;
BEGIN
  closing := TRUE
END StdCloseButtonCallback;

PROCEDURE SetStdCloseButtonCallback*;
BEGIN
  closing := FALSE;
  CloseButtonCallback := StdCloseButtonCallback
END SetStdCloseButtonCallback;

PROCEDURE Settings*(w, h: INTEGER; flags: SET);
BEGIN
  wantW := w; wantH := h;
  wantFullscreen := fullscreen IN flags;
  wantSpread := spread IN flags;
  wantMouse := initMouse IN flags;
  showMouse := wantMouse;
  wantNet := initNet IN flags;
  wantSound := initSound IN flags
END Settings;

PROCEDURE GetDesktopResolution*(VAR w, h: INTEGER);
VAR mode: SDL.DisplayMode;
BEGIN
  IF SDL.GetDesktopDisplayMode(0, mode) = 0 THEN END;
  w := mode.w; h := mode.h
END GetDesktopResolution;

PROCEDURE CheckWantedScreenSize;
BEGIN
  IF (wantW = -1) OR (wantH = -1) THEN
    GetDesktopResolution(wantW, wantH);
    IF ~wantFullscreen THEN
      DEC(wantW, 20); DEC(wantH, 50)
    END
  END
END CheckWantedScreenSize;

(* Flip Region *)
PROCEDURE SetRegion*(x, y, w, h: INTEGER);
BEGIN
  flipRegion.x := x;  flipRegion.y := y;
  flipRegion.w := w;  flipRegion.h := h
END SetRegion;

PROCEDURE UnsetRegion*;
BEGIN
  flipRegion.w := -1
END UnsetRegion;

PROCEDURE AddRegion*(x, y, w, h: INTEGER);
BEGIN
  IF flipRegion.w = -1 THEN (* No flip region yet *)
    SetRegion(x, y, w, h) (* Just set it *)
  ELSE (* Flip Region exists, add to it *)
    IF x < flipRegion.x THEN flipRegion.x := x END;
    IF y < flipRegion.y THEN flipRegion.y := y END;
    IF x + w > flipRegion.x + flipRegion.w THEN
      flipRegion.w := w + x - flipRegion.x END;
    IF y + h > flipRegion.y + flipRegion.h THEN
      flipRegion.h := h + y - flipRegion.y END
  END
END AddRegion;

(* Drawing *)

PROCEDURE FillRect0(surface: SDL.Surface; VAR rect: SDL.Rect; color: INTEGER);
BEGIN
  IF SDL.FillRect(surface, rect, color) = 0 THEN END
END FillRect0;

PROCEDURE FillRect1(surface: SDL.Surface; VAR rect: SDL.Rect; color: INTEGER);
VAR pad1: INTEGER;
BEGIN
  IF SDL.FillRect(surface, rect, color) = 0 THEN END
END FillRect1;

PROCEDURE FillRect2(surface: SDL.Surface; VAR rect: SDL.Rect; color: INTEGER);
VAR pad1, pad2: INTEGER;
BEGIN
  IF SDL.FillRect(surface, rect, color) = 0 THEN END
END FillRect2;

PROCEDURE FillRect3(surface: SDL.Surface; VAR rect: SDL.Rect; color: INTEGER);
VAR pad1, pad2, pad3: INTEGER;
BEGIN
  IF SDL.FillRect(surface, rect, color) = 0 THEN END
END FillRect3;

PROCEDURE FillRect(surface: SDL.Surface; VAR rect: SDL.Rect; color: INTEGER);
CONST SP = 4; (* Register number of stack pointer *)
VAR sp: INTEGER;
BEGIN
  CASE SYSTEM.ADR(sp) MOD 16 OF
    0: FillRect0(surface, rect, color)
  | 4: FillRect1(surface, rect, color)
  | 8: FillRect2(surface, rect, color)
  ELSE FillRect3(surface, rect, color) END
END FillRect;

PROCEDURE FillAll(surface: SDL.Surface; color: INTEGER);
VAR r: SDL.Rect;
BEGIN
  (* Equivalent to SDL.FillRect(surface, NIL, color) *)
  r.x := 0; r.y := 0; r.w := surface.w; r.h := surface.h;
  FillRect(surface, r, color)
END FillAll;

PROCEDURE RenderClear0(renderer: SDL.Renderer);
BEGIN
    IF SDL.RenderClear(renderer) = 0 THEN END
END RenderClear0;

PROCEDURE RenderClear1(renderer: SDL.Renderer);
VAR pad1: INTEGER;
BEGIN
    IF SDL.RenderClear(renderer) = 0 THEN END
END RenderClear1;

PROCEDURE RenderClear2(renderer: SDL.Renderer);
VAR pad1, pad2: INTEGER;
BEGIN
    IF SDL.RenderClear(renderer) = 0 THEN END
END RenderClear2;

PROCEDURE RenderClear3(renderer: SDL.Renderer);
VAR pad1, pad2, pad3: INTEGER;
BEGIN
    IF SDL.RenderClear(renderer) = 0 THEN END
END RenderClear3;

PROCEDURE RenderClear(renderer: SDL.Renderer);
VAR sp: INTEGER;
BEGIN
  CASE SYSTEM.ADR(sp) MOD 16 OF
    0: RenderClear0(renderer)
  | 4: RenderClear1(renderer)
  | 8: RenderClear2(renderer)
  ELSE RenderClear3(renderer) END
END RenderClear;

PROCEDURE GetPixels*(bmp: Bitmap; VAR pixels: SDL.Pixels);
BEGIN
  pixels := bmp.surface.pixels
END GetPixels;

PROCEDURE MakeCol*(r, g, b: INTEGER): INTEGER;
BEGIN
  r := SYSTEM.VAL(INTEGER, SYSTEM.VAL(SET, r) * {0..7});
  g := SYSTEM.VAL(INTEGER, SYSTEM.VAL(SET, g) * {0..7});
  b := SYSTEM.VAL(INTEGER, SYSTEM.VAL(SET, b) * {0..7});
  RETURN ASH(ASH(0FF00H + b, 8) + g, 8) + r
END MakeCol;

PROCEDURE ColorToRGB*(color: INTEGER; VAR r, g, b: INTEGER);
BEGIN
  r := SYSTEM.VAL(INTEGER, SYSTEM.VAL(SET, color) * {0..7});
  g := SYSTEM.VAL(INTEGER, SYSTEM.VAL(SET, ASH(color, -8)) * {0..7});
  b := SYSTEM.VAL(INTEGER, SYSTEM.VAL(SET, ASH(color, -16)) * {0..7})
END ColorToRGB;

PROCEDURE BmpCol*(bmp: Bitmap; r, g, b: INTEGER): INTEGER;
BEGIN
  RETURN SDL.MapRGB(bmp.surface.format, SHORT(CHR(r MOD 256)),
    SHORT(CHR(g MOD 256)), SHORT(CHR(b MOD 256)))
END BmpCol;

PROCEDURE ClearToColor* (bmp: Bitmap; color: INTEGER);
BEGIN
  FillAll(bmp.surface, color)
END ClearToColor;

PROCEDURE ClearBitmap* (bmp: Bitmap);
BEGIN
  FillAll(bmp.surface, MakeCol(0, 0, 0))
END ClearBitmap;

PROCEDURE ClearScreenToColor* (color: INTEGER);
BEGIN
  FillAll(screen.surface, color)
END ClearScreenToColor;

PROCEDURE ClearScreen*;
BEGIN
  FillAll(screen.surface, MakeCol(0, 0, 0))
END ClearScreen;

PROCEDURE ClearScreenToColor2* (s: Bitmap; color: INTEGER);
BEGIN
  ClearToColor(screen, color)
END ClearScreenToColor2;

PROCEDURE PutPixelQuick*(pixels: Pixels; w, x, y, color: INTEGER);
BEGIN
  pixels[y * w + x] := color
END PutPixelQuick;

PROCEDURE PutPixel*(bmp: Bitmap; x, y, color: INTEGER);
VAR pixels: Pixels;
BEGIN
  IF (x >= 0) & (x < bmp.w) &
     (y >= 0) & (y < bmp.h) THEN
    IF SDL.LockSurface(bmp.surface) = 0 THEN END;
    GetPixels(bmp, pixels);
    pixels[y * bmp.w + x] := color;
    SDL.UnlockSurface(bmp.surface)
  END
END PutPixel;

PROCEDURE GetPixel*(bmp: Bitmap; x, y: INTEGER): INTEGER;
VAR pixels: SDL.Pixels; color: INTEGER;
BEGIN
  IF (x >= 0) & (x < bmp.w) &
     (y >= 0) & (y < bmp.h) THEN
    IF SDL.LockSurface(bmp.surface) = 0 THEN END;
    GetPixels(bmp, pixels);
    color := pixels[y * bmp.w + x];
    SDL.UnlockSurface(bmp.surface)
  ELSE color := 0 END;
  RETURN color
END GetPixel;

PROCEDURE HLine*(bmp: Bitmap; x1, y, x2, color: INTEGER);
VAR rect: SDL.Rect; t: INTEGER;
BEGIN
  IF x1 > x2 THEN t := x1; x1 := x2; x2 := t END;
  rect.x := x1; rect.y := y;
  rect.w := x2 - x1 + 1; rect.h := 1;
  FillRect(bmp.surface, rect, color)
END HLine;

PROCEDURE VLine*(bmp: Bitmap; x, y1, y2, color: INTEGER);
VAR rect: SDL.Rect; t: INTEGER;
BEGIN
  IF y1 > y2 THEN t := y1; y1 := y2; y2 := t END;
  rect.x := x; rect.y := y1;
  rect.w := 1; rect.h := y2 - y1 + 1;
  FillRect(bmp.surface, rect, color)
END VLine;

PROCEDURE Line*(bmp: Bitmap; x1, y1, x2, y2, color: INTEGER);
VAR x, y, i, dx, dy, sx, sy, e: INTEGER; vert: BOOLEAN;
    pixels: Pixels;
BEGIN
  IF x1 = x2 THEN
  VLine(bmp, x1, y1, y2, color)
  ELSIF y1 = y2 THEN
  HLine(bmp, x1, y1, x2, color)
  ELSE
    IF SDL.LockSurface(bmp.surface) = 0 THEN END;
    GetPixels(bmp, pixels);
    dx := ABS(x1 - x2); dy := ABS(y1 - y2);
    IF x2 > x1 THEN sx := 1 ELSE sx := -1 END;
    IF y2 > y1 THEN sy := 1 ELSE sy := -1 END;
    x := x1; y := y1; vert := dy > dx;
    IF vert THEN i := dx; dx := dy; dy := i END;
    e := 2 * dy - dx;
    FOR i := 0 TO dx DO
      IF (x >= 0) & (x < bmp.w) &
         (y >= 0) & (y < bmp.h) THEN
        PutPixelQuick(pixels, bmp.w, x, y, color)
      END;
      IF e >= 0 THEN
        IF vert THEN INC(x, sx) ELSE INC(y, sy) END;
        DEC(e, 2 * dx)
      END;
      IF vert THEN INC(y, sy) ELSE INC(x, sx) END;
      INC(e, 2 * dy)
    END;
    SDL.UnlockSurface(bmp.surface)
  END
END Line;

PROCEDURE FastLine*(bmp: Bitmap; x1, y1, x2, y2, color: INTEGER);
BEGIN
  Line(bmp, x1, y1, x2, y2, color)
END FastLine;

PROCEDURE Rect*(bmp: Bitmap; x, y, w, h, color: INTEGER);
VAR rect: SDL.Rect;
BEGIN
  rect.x := x; rect.y := y; rect.w := 1; rect.h := h;
  FillRect(bmp.surface, rect, color);
  rect.x := x + w - 1;
  FillRect(bmp.surface, rect, color);
  rect.x := x; rect.w := w; rect.h := 1;
  FillRect(bmp.surface, rect, color);
  rect.y := y + h - 1;
  FillRect(bmp.surface, rect, color)
END Rect;

PROCEDURE RectFill*(bmp: Bitmap; x, y, w, h, color: INTEGER);
VAR rect: SDL.Rect;
BEGIN
  rect.x := x; rect.y := y; rect.w := w; rect.h := h;
  FillRect(bmp.surface, rect, color)
END RectFill;

PROCEDURE Circle*(bmp: Bitmap; x, y, r, color: INTEGER);
BEGIN
END Circle;

PROCEDURE CircleFill*(bmp: Bitmap; x, y, r, color: INTEGER);
BEGIN
END CircleFill;

PROCEDURE Ellipse*(bmp: Bitmap; x, y, rx, ry, color: INTEGER);
BEGIN
END Ellipse;

PROCEDURE EllipseFill*(bmp: Bitmap; x, y, rx, ry, color: INTEGER);
BEGIN
END EllipseFill;

PROCEDURE FloodFill*(bmp: Bitmap; x, y, color: INTEGER);
BEGIN
END FloodFill;

(* Bitmap *)

PROCEDURE CreateBitmap*(w, h: INTEGER): Bitmap;
VAR bmp: Bitmap;
BEGIN
  NEW(bmp);
  bmp.surface := SDL.CreateRGBSurface(0, w, h, 32,
    000000FFH, 0000FF00H, 00FF0000H,
    -01000000H);
  bmp.w := w;  bmp.h := h;
  RETURN bmp
END CreateBitmap;

PROCEDURE DestroyBitmap*(bmp: Bitmap);
BEGIN
  SDL.FreeSurface(bmp.surface)
END DestroyBitmap;

PROCEDURE LoadBitmap*(fname: ARRAY OF SHORTCHAR): Bitmap;
VAR bmp: Bitmap;
BEGIN
  NEW(bmp); bmp.surface := LoadBmp(fname);
  IF bmp.surface = NIL THEN bmp := NIL
  ELSE bmp.w := bmp.surface.w; bmp.h := bmp.surface.h END;
  RETURN bmp
END LoadBitmap;

PROCEDURE Blit*(src, dest: Bitmap; sx, sy, dx, dy, w, h: INTEGER);
VAR a, b: SDL.Rect;
BEGIN
  a.x := sx; a.y := sy; a.w := w; a.h := h;
  b.x := dx; b.y := dy;
  IF SDL.BlitSurface(src.surface, a, dest.surface, b) = 0 THEN END
END Blit;

PROCEDURE StretchBlit*(src, dest: Bitmap; sx, sy, sw, sh, dx, dy, w, h: INTEGER);
VAR a, b: SDL.Rect;
BEGIN
  a.x := sx; a.y := sy; a.w := sw; a.h := sh;
  b.x := dx; b.y := dy; b.w := w; b.h := h;
  IF SDL.BlitSurface(src.surface, a, dest.surface, b) = 0 THEN END
END StretchBlit;

PROCEDURE MaskedBlit*(src, dest: Bitmap; sx, sy, dx, dy, w, h: INTEGER);
BEGIN
  (*Al.MaskedBlit(src.bmp, dest.bmp, sx, sy, dx, dy, w, h)*)
END MaskedBlit;

PROCEDURE DrawSpriteEx*(dest, sprite: Bitmap; x, y, mode: INTEGER; flip: SET);
BEGIN
  (*Al.DrawSpriteEx(dest.bmp, sprite.bmp, x, y, mode, flip)*)
END DrawSpriteEx;

PROCEDURE DrawCharacterEx*(dest, sprite: Bitmap; x, y, color, bg: INTEGER);
BEGIN
  (*Al.DrawCharacterEx(dest.bmp, sprite.bmp, x, y, color, bg)*)
END DrawCharacterEx;

PROCEDURE SetColorKey*(bmp: Bitmap; color: INTEGER);
BEGIN
  IF SDL.SetColorKey(bmp.surface, 1, color) = 0 THEN END
END SetColorKey;

(* Font *)

PROCEDURE LoadFont*(fname: ARRAY OF SHORTCHAR; charW, charH: INTEGER): Font;
VAR bmp: Bitmap; font: Font;
BEGIN
  bmp := LoadBitmap(fname);
  IF bmp = NIL THEN font := NIL
  ELSE
    bmp.surface := SDL.ConvertSurface(bmp.surface,
      screen.surface.format,0);
    SetColorKey(bmp, BmpCol(bmp, 0, 0, 0));
    NEW(font); font.bmp := bmp;
    font.charW := charW; font.charH := charH;
    font.charsInRow := font.bmp.w DIV charW;
    font.charRows := font.bmp.h DIV charH
  END;
  RETURN font
END LoadFont;

PROCEDURE DrawCharacter*(dest: Bitmap; font: Font;
  x, y: INTEGER; ch: CHAR; fg: INTEGER);
VAR fx, fy, r, g, b: INTEGER; srcRect, dstRect: SDL.Rect;
BEGIN
  dstRect.x := x; dstRect.y := y;
  srcRect.x := ORD(ch) MOD font.charsInRow * font.charW;
  srcRect.y := ORD(ch) DIV font.charsInRow * font.charH;
  srcRect.w := font.charW; srcRect.h := font.charH;
  ColorToRGB(fg, r, g, b);
  IF SDL.SetSurfaceColorMod(font.bmp.surface,
    SHORT(CHR(r)), SHORT(CHR(g)), SHORT(CHR(b))) = 0 THEN END;
  IF SDL.BlitSurface(font.bmp.surface, srcRect,
    screen.surface, dstRect) = 0 THEN END
END DrawCharacter;

PROCEDURE DrawString*(dest: Bitmap; font: Font;
  x, y: INTEGER; s: ARRAY OF SHORTCHAR; fg: INTEGER);
VAR i, cx: INTEGER;
BEGIN
  i := 0; cx := x;
  WHILE (s[i] # 0X) & (cx < dest.w) DO
    DrawCharacter(dest, font, cx, y, s[i], fg);
    INC(i); INC(cx, font.charW)
  END
END DrawString;

(* Keyboard *)

PROCEDURE KeyPressed*(): BOOLEAN;
BEGIN
  RETURN keyLen > 0
END KeyPressed;

PROCEDURE ReadKey*(): INTEGER;
VAR key: INTEGER;
BEGIN
  IF keyLen > 0 THEN
    DEC(keyLen); key := keyBuf[keyLen]
  ELSE key := 0 END;
  RETURN key
END ReadKey;

PROCEDURE GetScancode*(key: INTEGER): INTEGER;
BEGIN
  RETURN ASH(key, -8)
END GetScancode;

PROCEDURE GetChar*(key: INTEGER): SHORTCHAR;
BEGIN
  RETURN SHORT(CHR(SYSTEM.VAL(INTEGER,
    SYSTEM.VAL(SET, key) * {0..7})));
END GetChar;

PROCEDURE GetKeyArray*(): KeyArray;
BEGIN
  RETURN SYSTEM.VAL(KeyArray, SDL.GetKeyboardState(NIL))
END GetKeyArray;

(*
PROCEDURE AltPressed*(): BOOLEAN;
VAR keys: KeyArray;
BEGIN
  (*keys := GetKeyArray();
  RETURN keys[keyAlt] OR keys[keyAltGr]*)
END AltPressed;

PROCEDURE ShiftPressed*(): BOOLEAN;
VAR keys: KeyArray;
BEGIN
  (*keys := GetKeyArray();
  RETURN keys[keyLShift] OR keys[keyRShift]*)
END ShiftPressed;

PROCEDURE ControlPressed*(): BOOLEAN;
VAR keys: KeyArray;
BEGIN
  (*keys := GetKeyArray();
  RETURN keys[keyLControl] OR keys[keyRControl]*)
END ControlPressed;
*)

(* Mouse *)

PROCEDURE MouseOnScreen*(): BOOLEAN;
VAR flags: SET;
BEGIN
  flags := SDL.GetWindowFlags(window);
  RETURN SDL.windowMouseFocus IN flags
END MouseOnScreen;

PROCEDURE ShowMouse*(show: BOOLEAN);
BEGIN
  showMouse := show
END ShowMouse;

PROCEDURE GetRealMousePos*(VAR x, y: INTEGER);
BEGIN
  IF SDL.GetMouseState(x, y) = 0 THEN END
END GetRealMousePos;

PROCEDURE GetMousePos*(VAR x, y: INTEGER);
BEGIN
  x := mouseX; y := mouseY
END GetMousePos;

PROCEDURE GetMouseButtons*(): SET;
VAR s: SET; x, y: INTEGER;
BEGIN
  s := SYSTEM.VAL(SET, SDL.GetMouseState(x, y));
  RETURN s
END GetMouseButtons;

PROCEDURE CreateStdMousePointer*(): Bitmap;
VAR bmp: Bitmap; fg, bg: INTEGER;
BEGIN
  bmp := CreateBitmap(12, 19);
  bg := MakeCol(255, 0, 255); fg := MakeCol(0, 0, 0);
  ClearToColor(bmp, bg);
  SetColorKey(bmp, bg);
  Line(bmp, 0, 0, 10, 10, fg);
  Line(bmp, 0, 0, 0, 14, fg);
  Line(bmp, 0, 14, 3, 11, fg); Line(bmp, 10, 10, 6, 10, fg);
  Line(bmp, 4, 12, 6, 17, fg); Line(bmp, 6, 11, 9, 17, fg);
  Line(bmp, 7, 18, 8, 18, fg);
  bg := MakeCol(255, 255, 255);
  VLine(bmp, 1, 2, 12, bg);
  VLine(bmp, 2, 3, 11, bg);
  VLine(bmp, 3, 4, 10, bg); VLine(bmp, 4, 5, 11, bg);
  VLine(bmp, 5, 6, 13, bg); VLine(bmp, 6, 7, 9, bg);
  VLine(bmp, 7, 8, 9, bg); VLine(bmp, 8, 9, 9, bg);
  VLine(bmp, 6, 12, 15, bg); VLine(bmp, 7, 14, 17, bg);
  VLine(bmp, 8, 16, 17, bg);
  RETURN bmp
END CreateStdMousePointer;

PROCEDURE SetMousePointer*(bmp: Bitmap);
BEGIN
  mousePointer := bmp;
  underMouse := CreateBitmap(mousePointer.w, mousePointer.h);
  needRedrawMouse := TRUE
END SetMousePointer;

PROCEDURE GetMousePointer*(): Bitmap;
BEGIN
  RETURN mousePointer
END GetMousePointer;

PROCEDURE SetMouseFocus*(x, y: INTEGER);
BEGIN
  mouseFocusX := x;  mouseFocusY := y;
  needRedrawMouse := TRUE
END SetMouseFocus;

PROCEDURE InitMouseData;
BEGIN
  SetMousePointer(CreateStdMousePointer());
  SetMouseFocus(0, 0)
END InitMouseData;

(* Misc *)
PROCEDURE SetWindowTitle*(title: ARRAY OF SHORTCHAR);
BEGIN
  SDL.SetWindowTitle(window, title)
END SetWindowTitle;

PROCEDURE SwitchToWindowed*;
VAR a: LONGINT;
BEGIN
  (*wantW := buffer.w; wantH := buffer.h;
  a := Al.SetGfxMode(Al.gfxAutodetectWindowed, wantW, wantH, 0, 0)*)
END SwitchToWindowed;

PROCEDURE SwitchToFullscreen*;
VAR a: LONGINT;
BEGIN
  (*a := Al.SetGfxMode(Al.gfxAutodetectFullscreen, wantW, wantH, wantW, wantH)*)
END SwitchToFullscreen;

PROCEDURE Rest*(n: INTEGER);
BEGIN
  SDL.Delay(n)
END Rest;

PROCEDURE SaveKeyInBuffer(VAR event: SDL.Event);
VAR key: INTEGER;
    e: SDL.KeyboardEvent;
    n: INTEGER;
BEGIN
  IF keyLen < LEN(keyBuf) THEN
    e := SYSTEM.VAL(SDL.KeyboardEvent, SYSTEM.ADR(event));
    key := ASH(e.keysym.scancode, 8) +
      SYSTEM.VAL(INTEGER,
        SYSTEM.VAL(SET, e.keysym.sym) * {0..7});
    keyBuf[keyLen] := key; INC(keyLen)
  END
END SaveKeyInBuffer;

PROCEDURE UpdateMousePos(event: SDL.Event);
VAR e: SDL.MouseMotionEvent; newX, newY: INTEGER;
BEGIN
  e := SYSTEM.VAL(SDL.MouseMotionEvent, SYSTEM.ADR(event));
  newX := e.x; newY := e.y;
  IF newX < 0 THEN newX := 0
  ELSIF newX >= screen.w THEN newX := screen.w - 1 END;
  IF newY < 0 THEN newY := 0
  ELSIF newY >= screen.h THEN newY := screen.h - 1 END;
  IF (newX # mouseX) OR (newY # mouseY) THEN
    mouseX := newX;  mouseY := newY;
    needRedrawMouse := TRUE
  END
END UpdateMousePos;

PROCEDURE Poll*;
VAR event: SDL.Event;
BEGIN
  WHILE SDL.PollEvent(event) = 1 DO
    CASE event.type OF
      SDL.mouseMotion: UpdateMousePos(event)
    | SDL.keyDown: SaveKeyInBuffer(event)
    | SDL.quit: CloseButtonCallback
    ELSE END
  END
END Poll;

PROCEDURE WindowShown*(): BOOLEAN;
VAR flags: SET;
BEGIN
  flags := SDL.GetWindowFlags(window);
  RETURN SDL.windowShown IN flags
END WindowShown;

PROCEDURE Flip*;
VAR mx, my: INTEGER; (* Mouse bitmap X Y *)
    blitMouse: BOOLEAN;
    screenTexture: SDL.Texture;

  PROCEDURE PrepareMouse;
  BEGIN
    mx := mouseX - mouseFocusX;
    my := mouseY - mouseFocusY;
    (* Save image under mouse from buffer *)
    Blit(screen, underMouse, mx, my,
      0, 0, underMouse.w, underMouse.h);
    (* Blit mouse pointer onto buffer *)
    Blit(mousePointer, screen, 0, 0,
      mx, my, mousePointer.w, mousePointer.h)
  END PrepareMouse;

  PROCEDURE CleanMouse;
  BEGIN
    (* Restore image under mouse in buffer *)
    Blit(underMouse, screen, 0, 0,
      mx, my, underMouse.w, underMouse.h);
    needRedrawMouse := FALSE
  END CleanMouse;

  PROCEDURE BlitMouseOutsideFlipRegion;
  BEGIN
    IF lastBlitMouseOutside THEN
      (*Blit(src, dest: Bitmap; sx, sy, dx, dy, w, h: LONGINT);
      Blit(screen, *)
      (*Al.Blit(buffer.bmp, Al.screen, lastBlitMouseX,
        lastBlitMouseY, lastBlitMouseX, lastBlitMouseY,
        mousePointer.w, mousePointer.h);*)
    END;
    IF blitMouse THEN
      IF (mx < flipRegion.x) OR (my < flipRegion.y) OR
         (mx + mousePointer.w > flipRegion.x + flipRegion.w) OR
         (my + mousePointer.h > flipRegion.y + flipRegion.h) THEN
        (*Al.Blit(buffer.bmp, Al.screen, mx, my, mx, my,
          mousePointer.w, mousePointer.h);*)
        lastBlitMouseOutside := TRUE;
        lastBlitMouseX := mx; lastBlitMouseY := my
      ELSE
        lastBlitMouseOutside := FALSE
      END
    END
  END BlitMouseOutsideFlipRegion;

BEGIN
  IF WindowShown() THEN
    mx := 0; my := 0;
    blitMouse := showMouse & MouseOnScreen();
    IF blitMouse THEN PrepareMouse END;

    (* Blit buffer on screen *)
    IF SDL.SetRenderDrawColor(renderer, 0X, 0X, 0X, 0FFX) = 0 THEN END;
  RenderClear(renderer);
  screenTexture := SDL.CreateTextureFromSurface(renderer,
    screen.surface);
    IF SDL.RenderCopy(renderer, screenTexture, NIL, NIL) = 0 THEN END;
    SDL.RenderPresent(renderer);
    SDL.DestroyTexture(screenTexture);
    IF blitMouse THEN CleanMouse END
  END;
  Poll
END Flip;

(* Sound *)
(*
PROCEDURE LoadSample*(fname: ARRAY OF CHAR): Sample;
VAR spl: Sample;
BEGIN
  NEW(spl);
  spl.chunk := Mix.LoadWavRW(SDL.RWFromFile(fname, 'rb'), 1);
  IF spl.chunk = NIL THEN spl := NIL END;
  RETURN spl
END LoadSample;

PROCEDURE PlayChannel*(spl: Sample; channel: INTEGER);
BEGIN
  IF spl # NIL THEN
    IF Mix.PlayChannelTimed(channel, spl.chunk, 0, -1) = 0 THEN END
  END
END PlayChannel;

PROCEDURE PlaySample*(spl: Sample);
BEGIN
  IF spl # NIL THEN PlayChannel(spl, -1) END
END PlaySample;

PROCEDURE StopSample*(spl: Sample);
BEGIN
  (*SDL.StopSample(spl)*)
END StopSample;

PROCEDURE LoadMusic*(fname: ARRAY OF SHORTCHAR): Music;
VAR mus: Music;
BEGIN
  NEW(mus); mus.music := Mix.LoadMus(fname);
  IF mus.music = NIL THEN mus := NIL END;
  RETURN mus
END LoadMusic;

PROCEDURE PlayMusic*(mus: Music);
BEGIN
  IF (mus # NIL) & (mus.music # NIL) THEN
    IF Mix.PlayMusic(mus.music, -1) = 0 THEN END
  END
END PlayMusic;

PROCEDURE StopMusic*;
BEGIN
  Mix.HaltMusic
END StopMusic;
*)
(* Net *)
(*
PROCEDURE OpenServer*(port: INTEGER): Socket;
VAR addr: IPAddress; socket: Socket;
BEGIN
  IF Net.ResolveHost(addr, NIL, SHORT(port)) = 0
  THEN socket := Net.TCPOpen(addr)
  ELSE socket := NIL END;
  RETURN socket
END OpenServer;

PROCEDURE CloseConn*(VAR socket: Socket);
BEGIN
  IF Net.TCPClose(socket) = 0 THEN socket := NIL END
END CloseConn;

PROCEDURE Accept*(server: Socket): Socket;
BEGIN
  RETURN Net.TCPAccept(server)
END Accept;

PROCEDURE ConnectTo*(ip: ARRAY OF SHORTCHAR; port: INTEGER): Socket;
VAR addr: IPAddress; socket: Socket;
BEGIN
  IF Net.ResolveHost(addr, ip, SHORT(port)) = 0
  THEN socket := Net.TCPOpen(addr)
  ELSE socket := NIL END;
  RETURN socket
END ConnectTo;

PROCEDURE SocketReady*(socket: Socket): BOOLEAN;
BEGIN
  RETURN (socket # NIL) & (socket.ready # 0)
END SocketReady;

PROCEDURE Send*(socket: Socket;
  data: ARRAY OF SHORTCHAR; len: INTEGER): INTEGER;
BEGIN
  RETURN Net.TCPSend(socket, data, len)
END Send;

PROCEDURE SendStr*(socket: Socket; s: ARRAY OF SHORTCHAR): INTEGER;
VAR len: LONGINT;
BEGIN len := 0;
  WHILE (len < LEN(s)) & (s[len] # 0X) DO INC(len) END;
  RETURN Net.TCPSend(socket, s, len)
END SendStr;

PROCEDURE Recv*(socket: Socket;
  VAR data: ARRAY OF SHORTCHAR; len: INTEGER): INTEGER;
BEGIN
  RETURN Net.TCPRecv(socket,
    SYSTEM.VAL(SYSTEM.PTR, SYSTEM.ADR(data[0])), len)
END Recv;

PROCEDURE RecvStr*(socket: Socket; VAR s: ARRAY OF SHORTCHAR): INTEGER;
VAR n: LONGINT;
BEGIN
  n := Recv(socket, s, LEN(s) - 1);
  IF n >= 0 THEN
    IF n >= LEN(s) THEN n := LEN(s) END;
    s[n] := 0X
  END;
  RETURN n
END RecvStr;
*)
(* Socket Set *)
(*
PROCEDURE AllocSocketSet*(maxSockets: INTEGER): SocketSet;
BEGIN
  RETURN Net.AllocSocketSet(maxSockets)
END AllocSocketSet;

PROCEDURE FreeSocketSet*(set: SocketSet);
BEGIN
  Net.FreeSocketSet(set)
END FreeSocketSet;

PROCEDURE AddSocket*(set: SocketSet; socket: Socket): INTEGER;
BEGIN
  RETURN Net.AddSocket(set, SYSTEM.VAL(Net.Socket, socket))
END AddSocket;

PROCEDURE DelSocket*(set: SocketSet; socket: Socket): INTEGER;
BEGIN
  RETURN Net.DelSocket(set, SYSTEM.VAL(Net.Socket, socket))
END DelSocket;

PROCEDURE CheckSockets*(set: SocketSet; timeout: INTEGER): BOOLEAN;
BEGIN
  RETURN Net.CheckSockets(set, timeout) # 0
END CheckSockets;
*)
(* NetBuf *)
(*
PROCEDURE InitNetBuf*(buf: NetBuf; size: INTEGER);
BEGIN
  NEW(buf.s, size); buf.len := 0; buf.lastLen := 0
END InitNetBuf;

PROCEDURE NewNetBuf*(size: INTEGER): NetBuf;
VAR buf: NetBuf;
BEGIN
  NEW(buf); InitNetBuf(buf, size); RETURN buf
END NewNetBuf;

PROCEDURE RecvBuf*(socket: Socket; buf: NetBuf);
VAR n, len: LONGINT; s: PStr;
BEGIN
  len := LEN(buf.s^) - buf.len;
  IF len = 0 THEN
    NEW(s, LEN(buf.s^) * 2); COPY(buf.s^, s^);
    buf.s := s
  END;
  n := Net.TCPRecv(socket,
    SYSTEM.VAL(SYSTEM.PTR, SYSTEM.ADR(buf.s[buf.len])),
    len);
  buf.lastLen := buf.len;
  IF n > 0 THEN
    INC(buf.len, n);
    IF buf.len > LEN(buf.s^) THEN buf.len := LEN(buf.s^) END
  END
END RecvBuf;

PROCEDURE BufGet*(buf: NetBuf; sep: SHORTCHAR;
  VAR s: ARRAY OF SHORTCHAR): BOOLEAN;
VAR i: SHORTINT; found: BOOLEAN;
BEGIN i := SHORT(buf.lastLen);
  WHILE (i < buf.len) & (buf.s[i] # sep) DO INC(i) END;
  found := i < buf.len;
  IF found THEN
    S.Extract(buf.s^, 0, i, s);
    S.Delete(buf.s^, 0, i + 1);
    DEC(buf.len, i + 1); buf.lastLen := 0
  END;
  RETURN found
END BufGet;
*)

PROCEDURE SpreadResolution;
VAR w, h, scale: INTEGER;
BEGIN
  GetDesktopResolution(w, h);
  IF w / h > wantW / wantH THEN 
    scale := h DIV wantH;
    IF scale > 1 THEN wantH := h DIV scale END; 
    wantW := w * wantH DIV h
  ELSE 
    scale := w DIV wantW;
    IF scale > 1 THEN wantW := w DIV scale END; 
    wantH := h * wantW DIV w
  END; 
  IF scale = 1 THEN
    IF SDL.SetHint(SDL.hintRenderScaleQuality, '1') = 0 THEN END
  END
END SpreadResolution;

(* Init *)

PROCEDURE Init*(): Bitmap;
VAR flags: SET; success: BOOLEAN; w, h: INTEGER;
BEGIN screen := NIL;
  IF SDL.Init({SDL.initVideo}) = 0 THEN
    flags := {};
    IF wantFullscreen THEN
      flags := flags + SDL.windowFullscreenDesktop;
      IF wantSpread THEN SpreadResolution END
    END;
    window := SDL.CreateWindow('',
      SDL.windowPosUndefined, SDL.windowPosUndefined,
      wantW, wantH, flags);
    IF window # NIL THEN
      renderer := SDL.CreateRenderer(window,
        -1, {SDL.rendererAccelerated});
    (* On Windows on VirtualBox use rendererSoftware instead *)
    IF renderer # NIL THEN
        IF SDL.RenderSetLogicalSize(renderer, wantW, wantH) = 0 THEN END;
        screen := CreateBitmap(wantW, wantH);
        SetStdCloseButtonCallback;
        UnsetRegion;
        IF SDL.ShowCursor(0) = 0 THEN END;
        IF wantMouse THEN InitMouseData END;
      (*
        IF wantSound THEN
          IF Mix.Init({}) = 0 THEN
            IF Mix.OpenAudio(22050, Mix.defaultFormat,
               4, 1024) = 0 THEN END
          END
      END;
        IF wantNet THEN
          IF Net.Init() = 0 THEN END
        END
      *)
    END
    END
  END;
  RETURN screen
END Init;

PROCEDURE Close*;
BEGIN
  (*
  IF wantSound THEN Mix.Quit END;
  IF wantNet THEN Net.Quit END;
  *)
  SDL.Quit
END Close;

BEGIN
  wantW := -1; wantH := -1; wantMouse := FALSE;
  wantFullscreen := TRUE; wantSpread := TRUE;
  mousePointer := NIL; lastBlitMouseOutside := FALSE;
  mouseFocusX := 0; mouseFocusY := 0
END GraphLib.

