MODULE GraphSdl ["libSDL2.so"];

IMPORT SYSTEM;

CONST
  (* SDL_Init Flags *)
  initTimer*          = 0;
  initAudio*          = 4;
  initVideo*          = 5;
  initJoystick*       = 9;
  initHaptic*         = 12;
  initGameController* = 13;
  initEvents*         = 14;
  initEverything*     = {initTimer, initAudio, initVideo, initJoystick,
                        initHaptic, initGameController, initEvents};

  (* SDL_CreateWindow Flags *)
  windowFullscreen*        = 0;
  windowOpenGL*            = 1;
  windowShown*             = 2;
  windowHidden*            = 3;
  windowBorderless*        = 4;
  windowResizable*         = 5;
  windowMinimized*         = 6;
  windowMaximized*         = 7;
  windowInputGrabbed*      = 8;
  windowInputFocus*        = 9;
  windowMouseFocus*        = 10;
  windowFullscreenDesktop* = {windowFullscreen, 12};
  windowForeign*           = 11;
  windowAllowHighDPI*      = 13;

  (* Window Position *)
  windowPosUndefined* = 1FFF0000H;

  (* Event Types *)
  quit*            = 100H;
  windowEvent*     = 200H;
  keyDown*         = 300H;
  keyUp*           = 301H;
  textEditing*     = 302H;
  textInput*       = 303H;
  mouseMotion*     = 400H;
  mouseButtonDown* = 401H;
  mouseButtonUp*   = 402H;
  mouseWheel*      = 403H;
  userEvent*       = 8000H;

  (* Key Mods *)
  kmodLShift*   = 0;
  kmodRShift*   = 1;
  kmodLCtrl*    = 6;
  kmodRCtrl*    = 7;
  kmodLAlt*     = 8;
  kmodRAlt*     = 9;
  kmodLGui*     = 10;
  kmodRGui*     = 11;
  kmodNum*      = 12;
  kmodCaps*     = 13;
  kmodMode*     = 14;
  kmodReserved* = 15;
  kmodCtrl*     = {kmodLCtrl, kmodRCtrl};
  kmodShift*    = {kmodLShift, kmodRShift};
  kmodAlt*      = {kmodLAlt, kmodRAlt};
  kmodGui*      = {kmodLGui, kmodRGui};

  (* Hints (for SDL_Hint) *)
  hintFramebufferAcceleration*  = 'SDL_FRAMEBUFFER_ACCELERATION';
  hintRenderDriver*             = 'SDL_RENDER_DRIVER';
  hintRenderOpenglShaders*      = 'SDL_RENDER_OPENGL_SHADERS';
  hintRenderDirect3dThreadsafe* = 'SDL_RENDER_DIRECT3D_THREAD_SAFE';
  hintRenderScaleQuality*       = 'SDL_RENDER_SCALE_QUALITY';
  hintRenderVsync*              = 'SDL_RENDER_VSYNC';
  hintVideoAllowScreensaver*    = 'SDL_VIDEO_ALLOW_SCREENSAVER';
  hintVideoX11Xvidmode*         = 'SDL_VIDEO_X11_XVIDMODE';
  hintVideoX11Xinerama*         = 'SDL_VIDEO_X11_XINERAMA';
  hintVideoX11Xrandr*           = 'SDL_VIDEO_X11_XRANDR';
  hintGrabKeyboard*             = 'SDL_GRAB_KEYBOARD';
  hintMouseRelativeModeWarp*    = 'SDL_MOUSE_RELATIVE_MODE_WARP';
  hintVideoMinimizeOnFocusLoss* = 'SDL_VIDEO_MINIMIZE_ON_FOCUS_LOSS';
  hintIdleTimerDisabled*        = 'SDL_IOSIDLETIMERDISABLED';
  hintOrientations*             = 'SDL_IOSORIENTATIONS'; (* !FIXME add _ lol *)
  hintAccelerometerAsJoystick*  = 'SDL_ACCELEROMETERASJOYSTICK';
  hintXinputEnabled*            = 'SDL_XINPUTENABLED';
  hintGamecontrollerconfig*     = 'SDL_GAMECONTROLLERCONFIG';
  hintJoystickAllowBackgroundEvents* = 'SDL_JOYSTICKALLOWBACKGROUNDEVENTS';
  hintAllowTopmost*                  = 'SDL_ALLOWTOPMOST';
  hintTimerResolution*               = 'SDL_TIMERRESOLUTION';
  hintVideoHighDpiDisabled*          = 'SDL_VIDEO_HIGHDPIDISABLED';
  hintMacCtrlClickEmulateRightClick* = 'SDL_MACCTRLCLICKEMULATERIGHTCLICK';
  hintVideoWinD3dCompiler*           = 'SDL_VIDEO_WIND3DCOMPILER';
  hintVideoWindowSharePixelFormat*   = 'SDL_VIDEO_WINDOWSHAREPIXELFORMAT';
  hintVideoMacFullscreenSpaces*      = 'SDL_VIDEO_MACFULLSCREENSPACES';

  (* Renderer Flags *)
  rendererSoftware*      = 0;
  rendererAccelerated*   = 1;
  rendererPresentVsync*  = 2;
  rendererTargetTexture* = 3;

  (* Flip Flags*)
  flipHorizontal* = 0;
  flipVertical*   = 1;

  (* ImgInit Flags *)
  imgInitJPG*  = 0;
  imgInitPNG*  = 1;
  imgInitTIF*  = 2;
  imgInitWEBP* = 3;

  (* Blend Modes *)
  blendModeBlend* = 0;
  blendModeAdd*   = 1;
  blendModeMod*   = 2;

  (* Mouse Buttons *)
  buttonLeft*      = 1;
  buttonMiddle*    = 2;
  buttonRight*     = 3;
  buttonWheelUp*   = 4;
  buttonWheelDown* = 5;
  buttonX1*        = 6;
  buttonX2*        = 7;

  (* Text Editing and Input *)
  textEditingEventTextSize* = 32;
  textInputEventTextSize*   = 32;

TYPE
  Ptr* = POINTER [untagged] TO RECORD [untagged] END;
  PStr* = POINTER [untagged] TO ARRAY [untagged] OF SHORTCHAR;

  Window* = Ptr;
  Renderer* = Ptr;
  Texture* = Ptr;

  Event* = RECORD [untagged]
    type*, timestamp*: INTEGER;
    padding: ARRAY [untagged] 48 OF SHORTCHAR
  END;
  
  WindowEvent* = POINTER [untagged] TO WindowEventDesc;
  WindowEventDesc* = RECORD [untagged]
    type*, timestamp*, windowID*: INTEGER;
    event*: SHORTCHAR;
    pad1, pad2, pad3: SHORTCHAR;
    data1*, data2*: INTEGER
  END;
  
  KeyboardEvent* = POINTER [untagged] TO KeyboardEventDesc;
  KeyboardEventDesc* = RECORD [untagged]
    type*, timestamp*, windowID*: INTEGER;
    state*, repeat*: SHORTCHAR;
    pad1, pad2: SHORTCHAR;
    keysym*: RECORD [untagged]
      scancode*, sym*: INTEGER;
      mod*: SHORTINT;
      unused: INTEGER
    END
  END;

  MouseMotionEvent* = POINTER [untagged] TO MouseMotionEventDesc;
  MouseMotionEventDesc* = RECORD [untagged]
    type*, timestamp*, windowID*: INTEGER;
    which*, state*: INTEGER;
    x*, y*: INTEGER;
    xrel*, yrel*: INTEGER
  END;

  MouseButtonEvent* = POINTER [untagged] TO MouseButtonEventDesc;
  MouseButtonEventDesc* = RECORD [untagged]
    type*, timestamp*, windowID*: INTEGER;
    which*: INTEGER;
    button*, state*, clicks*: SHORTCHAR;
    pad1: SHORTCHAR;
    x*, y*: INTEGER
  END;

  MouseWheelEvent* = POINTER [untagged] TO MouseWheelEventDesc;
  MouseWheelEventDesc* = RECORD [untagged]
    type*, timestamp*, windowID*: INTEGER;
    which*: INTEGER;
    x*, y*: INTEGER
  END;

  TextEditingEvent* = POINTER [untagged] TO TextEditingEventDesc;
  TextEditingEventDesc* = RECORD [untagged]
    type*, timestamp*, windowID*: INTEGER;
    text*: ARRAY [untagged] textEditingEventTextSize OF SHORTCHAR;
    start*, length*: INTEGER
  END;

  TextInputEvent* = POINTER [untagged] TO TextInputEventDesc;
  TextInputEventDesc* = RECORD [untagged]
    type*, timestamp*, windowID*: INTEGER;
    text*: ARRAY [untagged] textInputEventTextSize OF SHORTCHAR;
  END;

  Rect* = RECORD [untagged]
    x*, y*, w*, h*: INTEGER
  END;

  Palette* = POINTER [untagged] TO PaletteDesc;
  PaletteDesc* = RECORD [untagged]
    nColors*: INTEGER;
    colors*: Ptr;
    version*: INTEGER;
    refCount*: INTEGER
  END;

  DisplayMode* = RECORD [untagged]
    format*: INTEGER;
    w*, h*: INTEGER;
    refreshRate*: INTEGER;
    driverData: Ptr
  END;

  PixelFormat* = POINTER [untagged] TO PixelFormatDesc;
  PixelFormatDesc* = RECORD [untagged]
    format*: INTEGER;
    palette*: Palette;
    bitsPerPixel*, bytesPerPixel*: SHORTCHAR;
    pad1, pad2: SHORTCHAR;
    rMask*, gMask*, bMask*, aMask*: INTEGER;
    rLoss*, gLoss*, bLoss*, aLoss*: SHORTCHAR;
    rShift*, gShift*, bShift*, aShift*: SHORTCHAR;
    refCount*: INTEGER;
    next*: PixelFormat
  END;

  Pixels* = POINTER [untagged] TO ARRAY [untagged] OF INTEGER;

  Surface* = POINTER [untagged] TO SurfaceDesc;
  SurfaceDesc* = RECORD [untagged]
    flags: INTEGER;
    format*: PixelFormat;
    w*, h*, pitch*: INTEGER;
    pixels*: Pixels;
    userData*: Ptr;
    locked: INTEGER;
    lockData: Ptr;
    clipRect*: Rect;
    map: Ptr;
    refCount*: INTEGER
  END;

  KeyArray* = POINTER [untagged] TO ARRAY [untagged] 1024 OF BOOLEAN;

PROCEDURE [ccall] Init* ["SDL_Init"] (flags: SET): INTEGER;
PROCEDURE [ccall] Quit* ["SDL_Quit"] ;
PROCEDURE [ccall] Delay* ["SDL_Delay"] (ms: INTEGER);
PROCEDURE [ccall] CreateWindow* ["SDL_CreateWindow"] 
  (title: PStr; x, y, w, h: INTEGER; flags: SET): Window;
PROCEDURE [ccall] DestroyWindow* ["SDL_DestroyWindow"] (win: Window);
PROCEDURE [ccall] CreateRenderer* ["SDL_CreateRenderer"] 
  (win: Window; index: INTEGER; flags: SET): Renderer;
PROCEDURE [ccall] DestroyRenderer* ["SDL_DestroyRenderer"] (ren: Renderer);
PROCEDURE [ccall] RenderClear* ["SDL_RenderClear"] (ren: Renderer): INTEGER;
PROCEDURE [ccall] RenderPresent* ["SDL_RenderPresent"] (ren: Renderer);
PROCEDURE [ccall] SetRenderDrawColor* ["SDL_SetRenderDrawColor"]
  (ren: Renderer; r, g, b, a: SHORTCHAR): INTEGER;
PROCEDURE [ccall] PollEvent* ["SDL_PollEvent"] (VAR event: Event): INTEGER;
PROCEDURE [ccall] SetHint* ["SDL_SetHint"] (name, value: PStr): INTEGER;
PROCEDURE [ccall] RenderSetLogicalSize* ["SDL_RenderSetLogicalSize"]
  (ren: Renderer; w, h: INTEGER): INTEGER;
PROCEDURE [ccall] ShowCursor* ["SDL_ShowCursor"]
  (toggle: INTEGER): INTEGER;
PROCEDURE [ccall] CreateRGBSurface* ["SDL_CreateRGBSurface"]
  (flags, w, h, depth, rMask, gMask, bMask, aMask: INTEGER): Surface;
PROCEDURE [ccall] FreeSurface* ["SDL_FreeSurface"] (surface: Surface);
PROCEDURE [ccall] LockSurface* ["SDL_LockSurface"]
  (surface: Surface): INTEGER;
PROCEDURE [ccall] UnlockSurface* ["SDL_UnlockSurface"] (surface: Surface);
PROCEDURE [ccall] RenderCopy* ["SDL_RenderCopy"]
  (renderer: Renderer; texture: Texture;
  VAR [nil] srcRect, dstRect: Rect): INTEGER;
PROCEDURE [ccall] CreateTextureFromSurface*
  ["SDL_CreateTextureFromSurface"]
  (ren: Renderer; surface: Surface): Texture;
PROCEDURE [ccall] DestroyTexture* ["SDL_DestroyTexture"] (texture: Texture);
PROCEDURE [ccall] GetDesktopDisplayMode* ["SDL_GetDesktopDisplayMode"]
  (displayIndex: INTEGER; VAR mode: DisplayMode): INTEGER;
PROCEDURE [ccall] GetKeyboardState* ["SDL_GetKeyboardState"]
  (VAR [nil] numKeys: INTEGER): KeyArray;
PROCEDURE [ccall] GetMouseState* ["SDL_GetMouseState"]
  (VAR x, y: INTEGER): INTEGER;
PROCEDURE [ccall] FillRect* ["SDL_FillRect"]
  (dst: Surface; VAR [nil] rect: Rect; color: INTEGER): INTEGER;
PROCEDURE [ccall] RWFromFile* ["SDL_RWFromFile"]
  (fname, mode: PStr): Ptr;
PROCEDURE [ccall] LoadBmpRW* ["SDL_LoadBMP_RW"]
  (src: Ptr; freeSrc: INTEGER): Surface;
PROCEDURE [ccall] BlitSurface* ["SDL_UpperBlit"]
  (src: Surface; VAR [nil] srcRect: Rect;
  dst: Surface; VAR [nil] dstRect: Rect): INTEGER;
PROCEDURE [ccall] SetWindowTitle* ["SDL_SetWindowTitle"]
  (win: Window; title: PStr);
PROCEDURE [ccall] SetColorKey* ["SDL_SetColorKey"]
  (surface: Surface; flag, key: INTEGER): INTEGER;
PROCEDURE [ccall] SetSurfaceColorMod* ["SDL_SetSurfaceColorMod"]
  (surface: Surface; r, g, b: SHORTCHAR): INTEGER;
PROCEDURE [ccall] MapRGB* ["SDL_MapRGB"]
  (format: PixelFormat; r, g, b: SHORTCHAR): INTEGER;
PROCEDURE [ccall] ConvertSurface* ["SDL_ConvertSurface"]
  (src: Surface; fmt: PixelFormat; flags: INTEGER): Surface;
PROCEDURE [ccall] RenderSetScale* ["SDL_RenderSetScale"]
  (ren: Renderer; scaleX, scaleY: SHORTREAL): INTEGER;
PROCEDURE [ccall] GetError* ["SDL_GetError"] (): PStr;
PROCEDURE [ccall] GetWindowFlags* ["SDL_GetWindowFlags"]
  (win: Window): SET;

END GraphSdl.

